package JointFeldmanProtocol;

import Communication.User;
import FeldmanVerifiableSecretSharing.VerifiableSecretSharing;
import ShamirSecretSharing.Polynomial;
import com.google.protobuf.ByteString;
import meerkat.protobuf.DKGMessages;
import org.factcenter.qilin.primitives.Group;

import java.math.BigInteger;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

/**
 * Created by Tzlil on 3/14/2016.
 */
public class DistributedKeyGeneration extends VerifiableSecretSharing {
    public enum ComplainState{
        Non, Waiting,Disqualified,NonDisqualified
    }
    protected final int id;
    private DistributedKeyGenerationParty[] parties;


    public DistributedKeyGeneration(int t, int n, BigInteger zi, Random random, BigInteger q, BigInteger g
            , Group<BigInteger> group, int id)  {
        super(t, n, zi, random, q, g,group);
        this.id = id;
        this.parties = new DistributedKeyGenerationParty[n];
        for (int i = 1; i <= n ; i++){
            this.parties[i - 1] = new DistributedKeyGenerationParty(i,n,t);
        }
    }

    protected void setParties(DistributedKeyGenerationParty[] parties){
        this.parties = parties;
    }
    protected DistributedKeyGenerationParty[] getParties(){
        return parties;
    }

    /**
     * stage1.1 according to the protocol
     *  Pi broadcasts Aik for k = 0,...,t.
     */
    public void broadcastCommitments(User user){
        broadcastCommitments(user,commitmentsArray);
    }

    public void broadcastCommitments(User user, BigInteger[] commitments){
        DKGMessages.CommitmentMessage commitmentMessage;
        for (int k = 0; k <= t ; k++){
            commitmentMessage = DKGMessages.CommitmentMessage.newBuilder()
                    .setCommitment(ByteString.copyFrom(commitments[k].toByteArray()))
                    .setK(k)
                    .build();
            user.broadcast(DKGMessages.Mail.Type.COMMITMENT, commitmentMessage);
        }
    }

    public void sendSecret(User user, int j){
        ByteString secret = ByteString.copyFrom(getShare(j).y.toByteArray());
        user.send(j, DKGMessages.Mail.Type.SECRET,
                DKGMessages.SecretMessage.newBuilder()
                        .setI(id)
                        .setJ(j)
                        .setSecret(secret)
                        .build());
    }

    /**
     * stage1.2 according to the protocol
     *  Pi computes the shares Sij for j = 1,...,n and sends Sij secretly to Pj.
     */
    public void sendSecrets(User user){
        for (int j = 1; j <= n ; j++){
            if(j != id){
                sendSecret(user,j);
            }
        }
    }

    public boolean isValidSecret(int i){
        DistributedKeyGenerationParty party = parties[i - 1];
        return isValidSecret(party.share,party.commitments,id);
    }

    public boolean isValidSecret(Polynomial.Point secret, BigInteger[] commitments, int j){
        BigInteger v = verify(j,commitments,group);
        return group.multiply(g,secret.y).equals(v);
    }

    /**
     * stage2 according to the protocol
     *  Pj verifies all the shares he received (using isValidSecret)
     *  if check fails for an index i, Pj broadcasts a complaint against Pi.
     */
    public void broadcastComplains(User user){
        DKGMessages.IDMessage complaint;
        for (int i = 1; i <= n ; i++ ){
            if(i != id && !parties[i - 1].aborted) {
                sendComplain(user,i);
            }
        }
    }

    protected void sendComplain(User user,int i){
        DKGMessages.IDMessage complaint;
        if (!isValidSecret(i)) {
            //message = new Message(Type.Complaint, j)
            complaint = DKGMessages.IDMessage.newBuilder()
                    .setId(i)
                    .build();
            user.broadcast(DKGMessages.Mail.Type.COMPLAINT, complaint);
        }
    }


    public void broadcastComplaintAnswer(User user, int j){
        user.broadcast(DKGMessages.Mail.Type.ANSWER, DKGMessages.SecretMessage.newBuilder()
                .setI(id)
                .setJ(j)
                .setSecret(ByteString.copyFrom(getShare(j).y.toByteArray()))
                .build());
    }

    /**
     * stage3.1 according to the protocol
     *   if more than t players complain against a player Pi he is disqualified.
     */
    public void answerAllComplainingPlayers(User user){
        ComplainState[] complains = parties[id - 1].complaints;
        for (int i = 1; i <= n ; i++) {
            switch (complains[i - 1]) {
                case Waiting:
                    broadcastComplaintAnswer(user,i);
                    break;
                default:
                    break;
            }
        }
    }

    protected boolean isPartyCompletedStage1(int i){
        if(parties[i - 1].aborted){
            if(parties[i - 1].share == null){
                return false;
            }
            for (int k = 0; k <= t ; k++){
                if(parties[i - 1].commitments[k] == null){
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * stage3.2 according to the protocol
     *  if any of the revealed shares fails the verification test, player Pi is disqualified.
     *  set QUAL to be the set of non-disqualified players.
     */
    public Set<Integer> calcQUAL(){
        Set<Integer> QUAL = new HashSet<Integer>();
        boolean nonDisqualified;
        int counter;
        for (int i = 1; i <= n; i++){
            ComplainState[] complains = parties[i - 1].complaints;
            nonDisqualified = true;
            counter = 0;
            for (int j = 1; j <= n; j++){
                switch (complains[j - 1]) {
                    case Non:
                        break;
                    case NonDisqualified:
                        counter++;
                    default:
                        nonDisqualified = false;
                }
                if(!nonDisqualified)
                    break;
            }
            if(nonDisqualified && counter <= t && isPartyCompletedStage1(i)){
                QUAL.add(i);
            }
        }
        return QUAL;
    }

    /**
     * stage4.1 according to the protocol
     *   public value y is computed as y = multiplication of yi mod p for i in QUAL
     */
    public BigInteger calcY(Set<Integer> QUAL){
        BigInteger y = group.zero();
        for (int i : QUAL) {
            y = group.add(y , parties[i - 1].commitments[0]);
        }
        return y;
    }

    /**
     * stage4.2 according to the protocol
     *  public verification values are computed as Ak = multiplication of Aik mod p for i in QUAL for k = 0,...,t
     */
    public BigInteger[] calcCommitments(Set<Integer> QUAL){
        BigInteger[] commitments = new BigInteger[t + 1];
        Arrays.fill(commitments,group.zero());
        for (int i : QUAL) {
            for (int k = 0; k <= t; k++){
                commitments[k] = group.add(commitments[k], parties[i - 1].commitments[k]);
            }
        }
        return commitments;
    }

    /**
     * stage4.3 according to the protocol
     *  Pj sets is share of the secret as xj = sum of Sij mod q for i in QUAL
     */
    public Polynomial.Point calcShare(Set<Integer> QUAL){
        BigInteger xj = BigInteger.ZERO;
        for (int i : QUAL) {
            xj = xj.add(parties[i - 1].share.y);
        }
        return new Polynomial.Point(BigInteger.valueOf(id) , xj.mod(q));
    }

    public int getId() {
        return id;
    }

}
